"""
The filters.
"""

from ..error import TeapotError
from ..log import Highlight as hl
from .filter import Filter, f, uf


class FilteredObject(object):
    """
    Inherit from this class when your object can be disabled by filters.
    """

    def __init__(self, filter=None, *args, **kwargs):
        super(FilteredObject, self).__init__(*args, **kwargs)
        self._filter = filter

    @classmethod
    def get_enabled_instances(cls, keys_list=None, instances=None):
        if instances is None:
            instances = cls.get_instances(keys_list=keys_list)

        return [x for x in instances if x.enabled]

    @property
    def filter(self):
        if isinstance(self._filter, basestring):
            return Filter.get_instance_or_fail(self._filter)

        return self._filter

    @filter.setter
    def filter(self, filter):
        """
        Set the `filter` to check.

        `filter` can be None, a string, or a Filter instance.
        """

        self._filter = filter

        return self

    @property
    def enabled(self):
        """
        Check if the instance is filtered out.
        """

        return self.filter if self.filter is not None else True



"""
A source class.
"""

from .filters import FilteredObject
from .fetchers import Fetcher
from .memoized import MemoizedObject


class Source(MemoizedObject, FilteredObject):

    """
    Represents a project to build.
    """

    memoization_keys = ('attendee', 'resource')
    propagate_memoization_keys = True

    @classmethod
    def transform_memoization_keys(cls, attendee, resource):
        """
        Make sure the attendee parameter is a real Attendee instance.
        """

        if isinstance(attendee, basestring):
            from .attendee import Attendee

            attendee = Attendee(attendee)

        return attendee, resource

    def __init__(self, attendee, resource, mimetype=None, fetcher=None, *args, **kwargs):
        """
        Create a source that maps on the specified resource.
        """
        super(Source, self).__init__(*args, **kwargs)

        self.mimetype = mimetype
        self._fetcher = fetcher
        self._parsed_source = None

        # Register the source in the Attendee.
        self.attendee = attendee
        self.resource = resource
        attendee.add_source(self)

    def __str__(self):
        """
        Get a string representation of the source.
        """

        return self.resource

    @property
    def fetcher(self):
        if not self._fetcher:
            self._fetcher = Fetcher.get_instance_for(source=self)

        if isinstance(self._fetcher, basestring):
            self._fetcher = Fetcher.get_instance_or_fail(name=self._fetcher)

        return self._fetcher

    @property
    def parsed_source(self):
        if self._parsed_source is None:
            self._parsed_source = self.fetcher.parse_source(source=self)

        return self._parsed_source

    def fetch(self, target_path):
        """
        Fetches the source.
        """

        return self.fetcher.fetch(
            parsed_source=self.parsed_source,
            target_path=target_path,
        )



"""
A base Fetcher class.
"""

from ..memoized import MemoizedObject
from ..log import Highlight as hl

from .callbacks import ProgressBarFetcherCallback


class FetcherImplementation(object):

    """
    Base class for all fetcher implementation classes.
    """

    def parse_source(self, source):
        """
        Reimplement this method to indicate whether or not your fetcher class
        supports the specified `source`.

        Upon success, this method should return a truthy value if the specified
        `source` is supported by the fetcher. This truthy value will be passed
        as the `fetch_info` attribute to fetch().
        """

        raise NotImplementedError

    def fetch(self, fetch_info, target_path, progress):
        """
        Reimplement this method with your specific fetcher logic.

        `source` is the source to fetch.
        `target_path` is the path to the target.
        `progress` is an instance of a fetcher callback used to report progress
        to the user.

        This method must return the download manifest.

        It must raise an exception on error.

        You can (and should) provide feedback on the fetching operation by
        calling `self.progress.on_start`, `self.progress.on_update` and
        `self.progress.on_finish` at the appropriate time.

        See :class:`teapot.fetchers.callbacks.BaseFetcherCallback`
        documentation for details.
        """

        raise NotImplementedError


class Fetcher(MemoizedObject):

    """
    Represent a fetcher.
    """

    @classmethod
    def get_instance_for(cls, source, default=None):
        for instance in cls.get_instances():
            if instance.parse_source(source):
                return instance

        return default

    def __init__(self, fetcher_impl_class, progress_class=ProgressBarFetcherCallback):
        """
        Create a new fetcher that uses the specified fetcher
        implementation.
        """

        self._fetcher_impl = fetcher_impl_class()
        self.progress = progress_class()

    def parse_source(self, source):
        """
        Parse the specified source and return the parsed source.
        """

        return self._fetcher_impl.parse_source(source=source)

    def fetch(self, parsed_source, target_path):
        """
        Fetch the specified `parsed_source` using `target_path` as the target path.

        If fetching is not supported, a falsy value is returned.
        """

        try:
            return self._fetcher_impl.fetch(
                fetch_info=parsed_source,
                target_path=target_path,
                progress=self.progress
            )

        except Exception as ex:
            self.progress.on_exception(ex)

            raise


class register_fetcher(object):
    """
    A decorator that registers a fetcher.
    """

    def __init__(self, name):
        self.name = name

    def __call__(self, cls):
        with Fetcher.raise_on_duplicate():
            Fetcher(name=self.name, fetcher_impl_class=cls)

        return cls



"""
teapot unit tests.
"""

import os
import sys

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from teapot import *
from teapot.memoized import Memoized
from teapot.extensions import parse_extension
from teapot.error import TeapotError


class TestTeapot(unittest.TestCase):

    """
    Tests the different teapot components.
    """

    def setUp(self):
        """
        Clears all memoized instances.
        """

        Memoized.clear_all_instances()

    def test_environments(self):
        """
        Test the environments.
        """

        os.environ['DUMMY'] = 'DUMMY1'
        os.environ['FOO'] = 'FOO1'
        os.environ['HELLO'] = 'HELLO1'

        # We test the variables before we enable the environment
        self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
        self.assertEqual(os.environ.get('FOO'), 'FOO1')
        self.assertEqual(os.environ.get('BAR'), None)
        self.assertEqual(os.environ.get('HELLO'), 'HELLO1')

        # The different environments to test.
        empty_environment = Environment('empty')
        system_environment = Environment('system', variables=os.environ.copy())

        # We test that no variable are propagated inside the empty environment.
        with empty_environment.enable():
            self.assertEqual(os.environ.get('DUMMY'), None)
            self.assertEqual(os.environ.get('FOO'), None)
            self.assertEqual(os.environ.get('BAR'), None)
            self.assertEqual(os.environ.get('HELLO'), None)

        # We test that variables within the system environment are the same
        # than before.
        with system_environment.enable():
            self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
            self.assertEqual(os.environ.get('FOO'), 'FOO1')
            self.assertEqual(os.environ.get('BAR'), None)
            self.assertEqual(os.environ.get('HELLO'), 'HELLO1')

        self.assertEqual(system_environment.shell, None)

        environment = Environment(
            name='test_environment',
            variables={
                'FOO': 'FOO2',
                'BAR': 'BAR1',
                'HELLO': None,
                'FOO_EXTENDED': '$FOO-$FOO-$NON_EXISTING_VAR-$FOO_EXTENDED',
                'FOO_EXTENDED_PLATFORM': '[%HELLO%]',
                'NON_EXISTING': None,
            },
            parent=system_environment,
            shell=['my shell'],
        )

        # We test the variables before we enable the environment
        self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
        self.assertEqual(os.environ.get('FOO'), 'FOO1')
        self.assertEqual(os.environ.get('BAR'), None)
        self.assertEqual(os.environ.get('HELLO'), 'HELLO1')
        self.assertEqual(os.environ.get('FOO_EXTENDED'), None)
        self.assertTrue('NON_EXISTING' not in os.environ)

        # We apply the environment and test those again
        with environment.enable():
            self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
            self.assertEqual(os.environ.get('FOO'), 'FOO2')
            self.assertEqual(os.environ.get('BAR'), 'BAR1')
            self.assertEqual(os.environ.get('HELLO'), None)
            self.assertEqual(os.environ.get('FOO_EXTENDED'), 'FOO1-FOO1--')
            self.assertTrue('NON_EXISTING' not in os.environ)

            if sys.platform.startswith('win32'):
                self.assertEqual(os.environ.get('FOO_EXTENDED_PLATFORM'), '[HELLO1]')
            else:
                self.assertEqual(os.environ.get('FOO_EXTENDED_PLATFORM'), '[%HELLO%]')

        self.assertEqual(environment.shell, ['my shell'])

        sub_environment = Environment(
            name='test_environment_sub',
            parent=environment,
            shell=True,
        )

        self.assertEqual(sub_environment.shell, ['my shell'])

        orphan_environment = Environment(
            name='orphan_environment',
            variables={
                'FOO': 'FOO3',
                'BAR': 'BAR1',
                'HELLO': None,
                'NON_EXISTING': None,
            },
            parent=None,
        )

        # We test the variables before we enable the environment
        self.assertEqual(os.environ.get('DUMMY'), 'DUMMY1')
        self.assertEqual(os.environ.get('FOO'), 'FOO1')
        self.assertEqual(os.environ.get('BAR'), None)
        self.assertEqual(os.environ.get('HELLO'), 'HELLO1')
        self.assertTrue('NON_EXISTING' not in os.environ)

        # We apply the environment and test those again
        with orphan_environment.enable():
            self.assertEqual(os.environ.get('DUMMY'), None)
            self.assertEqual(os.environ.get('FOO'), 'FOO3')
            self.assertEqual(os.environ.get('BAR'), 'BAR1')
            self.assertEqual(os.environ.get('HELLO'), 'HELLO1')
            self.assertTrue('NON_EXISTING' not in os.environ)

        self.assertEqual(orphan_environment.shell, None)

        shell_environment = Environment(
            name='shell_environment',
            variables={
                'FOO': 'FOO3',
            },
            parent=system_environment,
            shell=['$FOO'],
        )

        self.assertEqual(shell_environment.shell, ['FOO1'])

        orphan_shell_environment = Environment(
            name='orphan_shell_environment',
            variables={
                'FOO': 'FOO3',
            },
            shell=['$FOO'],
        )

        self.assertEqual(orphan_shell_environment.shell, [''])

        # Test signatures changes

        signature = environment.signature
        sub_signature = sub_environment.signature

        environment.variables['NEWVARIABLE'] = 'bar'

        self.assertNotEqual(signature, environment.signature)
        self.assertNotEqual(sub_environment, sub_environment.signature)

        signature = environment.signature
        sub_signature = sub_environment.signature

        environment.shell = ['foo', 'goo']

        self.assertNotEqual(signature, environment.signature)
        self.assertNotEqual(sub_environment, sub_environment.signature)

        signature = environment.signature
        sub_signature = sub_environment.signature

        sub_environment.variables['NEWVARIABLE'] = 'bar'

        self.assertEqual(signature, environment.signature)
        self.assertNotEqual(sub_environment, sub_environment.signature)

        signature = environment.signature
        sub_signature = sub_environment.signature

        sub_environment.shell = ['foo', 'goo', 'boo']

        self.assertEqual(signature, environment.signature)
        self.assertNotEqual(sub_environment, sub_environment.signature)

    def test_filters(self):
        """
        Test the filters.
        """

        f('true', condition=True)
        f('false', condition=False)

        self.assertTrue(f('true'))
        self.assertFalse(f('false'))

        FUNCTION_RESULT = True

        def function():
            return FUNCTION_RESULT

        f('function', condition=function)

        self.assertTrue(f('function'))

        FUNCTION_RESULT = False

        self.assertFalse(f('function'))
        self.assertTrue(f('true') | f('false'))
        self.assertFalse(f('true') & f('false'))
        self.assertFalse(~f('true'))
        self.assertTrue(f('true') ^ f('false'))
        self.assertFalse(f('true') ^ f('true'))
        self.assertFalse(f('false') ^ f('false'))

    def test_attendees(self):
        """
        Test the attendees.
        """

        a = Attendee('a')
        b = Attendee('b')
        c = Attendee('c')
        d = Attendee('d')

        self.assertEqual(a, Attendee('a'))

        a.depends_on('b', c)
        c.depends_on('d')

        self.assertEqual(a.parents, {b, c})
        self.assertEqual(b.parents, set())
        self.assertEqual(c.parents, {d})
        self.assertEqual(d.parents, set())
        self.assertEqual(a.children, set())
        self.assertEqual(b.children, {a})
        self.assertEqual(c.children, {a})
        self.assertEqual(d.children, {c})

        self.assertEqual(
            Attendee.get_dependent_instances(),
            [b, d, c, a],
        )

        self.assertEqual(
            Attendee.get_dependent_instances(['c']),
            [d, c],
        )

        # Create a circular dependency.
        d.depends_on(a)

        self.assertRaises(Attendee.DependencyCycleError, Attendee.get_dependent_instances)

        # Add a source.
        a.add_source('http://some.fake.address')
        self.assertIsNotNone(a.get_source('http://some.fake.address'))

        # Add a build.
        attendee_test_environment = Environment('attendee_test_environment')

        a.add_build('foo', environment='attendee_test_environment')
        self.assertIsNotNone(a.get_build('foo'))
        self.assertEqual(a.get_build('foo').environment, attendee_test_environment)

    def test_extensions(self):
        """
        Test the extensions.
        """

        simple_call = 'foo'
        function_call = 'foo()'
        param_call = 'foo(1, 2, 3)'
        named_param_call = 'foo(a =1,c= 3,b = 2)'
        mixed_call = 'foo ( 1 , c=3, b = "test" )'
        missing_call = 'bar(5, 6, 7)'

        @register_extension('foo')
        def foo(a=None, b=None, c=None):
            return [a, b, c]

        extension, args, kwargs = parse_extension(simple_call)

        self.assertEqual(args, tuple())
        self.assertEqual(kwargs, {})

        extension, args, kwargs = parse_extension(function_call)

        self.assertEqual(args, tuple())
        self.assertEqual(kwargs, {})

        extension, args, kwargs = parse_extension(param_call)

        self.assertEqual(args, (1, 2, 3))
        self.assertEqual(kwargs, {})

        extension, args, kwargs = parse_extension(named_param_call)

        self.assertEqual(args, tuple())
        self.assertEqual(kwargs, {'a': 1, 'b': 2, 'c': 3})

        extension, args, kwargs = parse_extension(mixed_call)

        self.assertEqual(args, (1,))
        self.assertEqual(kwargs, {'b': 'test', 'c': 3})

        # Make sure unregistered extensions don't parse successfully.
        self.assertRaises(TeapotError, parse_extension, missing_call)


if __name__ == '__main__':
    unittest.main()



"""
An unpacker that deals with .zip files.
"""

from teapot.unpackers.unpacker import UnpackerImplementation
from teapot.unpackers.unpacker import register_unpacker

from ..error import TeapotError
from ..log import Highlight as hl

import os
import zipfile


@register_unpacker(('application/zip', None))
class ZipFileUnpacker(UnpackerImplementation):

    """
    An unpacker class that deals with .zip files.
    """

    def unpack(self, archive_path, target_path, progress):
        """
        Uncompress the archive.
        """

        if not zipfile.is_zipfile(archive_path):
            raise TeapotError(
                "%s is not a valid zip archive.",
                hl(archive_path),
            )

        zipf = zipfile.ZipFile(archive_path, 'r')

        # We get the common prefix for all archive members.
        prefix = os.path.commonprefix(zipf.namelist())

        while prefix and not prefix.endswith('/'):
            prefix = os.path.dirname(prefix)

        # We remove the trailing '/'
        prefix = os.path.dirname(prefix)

        if not prefix:
            raise TeapotError(
                "Unable to find a common prefix in %s to extract from.",
                hl(archive_path),
            )

        extracted_sources_path = os.path.join(target_path, prefix)

        progress.on_start(archive_path=archive_path, count=len(zipf.infolist()))

        for index, info in enumerate(zipf.infolist()):
            if os.path.isabs(info.filename):
                raise ValueError('Refusing to extract archive that contains absolute filenames.')

            progress.on_update(current_file=info.filename, progress=index)
            zipf.extract(info, path=target_path)

        progress.on_finish()

        return {
            'extracted_sources_path': extracted_sources_path,
        }



"""
The built-in filters.
"""

import os
import sys
import distutils
import subprocess

from .filter import named_filter


@named_filter('windows')
def windows():
    """
    Check if the platform is windows.
    """

    return sys.platform.startswith('win32')


@named_filter('linux')
def linux():
    """
    Check if the platform is windows.
    """

    return sys.platform.startswith('linux')


@named_filter('darwin')
def darwin():
    """
    Check if the platform is darwin.
    """

    return sys.platform.startswith('darwin')


@named_filter('unix')
def unix():
    """
    Check if the platform is unix.
    """

    return linux() or darwin()


@named_filter('msvc', depends_on='windows')
def msvc():
    """
    Check if MSVC is available.
    """

    return 'VCINSTALLDIR' in os.environ


@named_filter('msvc-x86', depends_on='msvc')
def msvc_x86():
    """
    Check if MSVC x86 is available.
    """

    output = subprocess.check_output(
        'cl.exe', shell=True, stderr=subprocess.STDOUT)
    first_line = output.split('\n')[0].rstrip()

    return 'x86' in first_line


@named_filter('msvc-x64', depends_on='msvc')
def msvc_x64():
    """
    Check if MSVC x64 is available.
    """

    output = subprocess.check_output(
        'cl.exe', shell=True, stderr=subprocess.STDOUT)
    first_line = output.split('\n')[0].rstrip()

    return 'x64' in first_line


@named_filter('mingw', depends_on='windows')
def mingw():
    """
    Check if MinGW is available.
    """

    return bool(distutils.spawn.find_executable('gcc'))



"""
A base class for object that have a signature.
"""

import hashlib


class SignableObject(object):

    """
    A base, signable object.
    """

    signature_fields = tuple()

    def __init__(self, *args, **kwargs):
        """
        Get the signable object instance.
        """
        super(SignableObject, self).__init__()

        self.hash = hashlib.sha1

    def get_signature_for(self, value):
        """
        Get the signature for a given value.
        """

        if isinstance(value, SignableObject):
            return value.signature

        m = self.hash()

        if isinstance(value, (list, tuple, set)):
            m.update('begin_list')
            map(m.update, map(self.get_signature_for, value))
            m.update('end_list')
        elif isinstance(value, dict):
            m.update('begin_dict')
            for key, value in value.iteritems():
                m.update(self.get_signature_for(key))
                m.update(self.get_signature_for(value))
            m.update('end_dict')
        elif value is None:
            m.update('none_value')
        elif value is True:
            m.update('true_value')
        elif value is False:
            m.update('false_value')
        else:
            m.update('begin_value')
            m.update(str(value))
            m.update('end_value')

        return m.hexdigest()

    @property
    def signature(self):
        return self.get_signature_for(map(
            lambda field: getattr(self, field),
            self.signature_fields
        ))



"""
A prefix class.
"""

import os


class PrefixedObject(object):

    """
    A class that has a prefix member.
    """

    def __init__(self, prefix=None, *args, **kwargs):
        super(PrefixedObject, self).__init__(*args, **kwargs)
        self._prefix = prefix

    @property
    def prefix(self):
        return self._prefix or ''

    @prefix.setter
    def prefix(self, value):
        self._prefix = value



"""
An attendee class.
"""

import os
import json
import hashlib

from .memoized import MemoizedObject
from .filters import FilteredObject
from .source import Source
from .error import TeapotError
from .log import LOGGER, Highlight as hl
from .options import get_option
from .path import mkdir, rmdir, from_user_path, temporary_copy
from .unpackers import Unpacker
from .build import Build
from .globals import get_party_path
from .prefix import PrefixedObject


class Attendee(MemoizedObject, FilteredObject, PrefixedObject):

    """
    Represents a project to build.
    """

    class DependencyCycleError(TeapotError):
        def __init__(self, cycle):
            cycle_str = ' -> '.join(map(str, cycle))

            super(Attendee.DependencyCycleError, self).__init__(
                'Dependency cycle found: %s',
                hl(cycle_str)
            )

            self.cycle = cycle

    @classmethod
    def get_dependent_instances(cls, keys_list=None):
        """
        Get the dependent instances, less dependent instances first.
        """

        try:
            instances = cls.get_enabled_instances(keys_list=keys_list)
        except Attendee.NoSuchInstance as ex:
            raise TeapotError(
                (
                    "Reference to a non-existing attendee %s could not be "
                    "resolved."
                ),
                hl(ex.keys[0]),
            )

        result = []
        dependency_tree = {
            attendee: attendee.parents
            for attendee in cls.get_enabled_instances()
        }

        while dependency_tree:
            try:
                attendee = next(x[0] for x in dependency_tree.iteritems() if not x[1])
            except StopIteration:
                # We can't find an attendee that does not depend on any
                # other. We have a dependency cycle.
                cycle = []
                attendee = dependency_tree.keys()[0]

                while not attendee in cycle:
                    cycle.append(attendee)
                    attendee = next(iter(dependency_tree[attendee]))

                # We make sure the dependency cycle starts with the
                # first repeating element (which is currently in
                # `attendee`)
                cycle = cycle[cycle.index(attendee):] + [attendee]

                raise Attendee.DependencyCycleError(cycle)

            else:
                result.append(attendee)
                del dependency_tree[attendee]

                dependency_tree = {
                    key: value - {attendee}
                    for key, value in dependency_tree.iteritems()
                }

        unneeded_results = set(result)

        def filter_out(instance):
            if instance in unneeded_results:
                unneeded_results.remove(instance)
                map(filter_out, instance.parents)

        map(filter_out, instances)

        return [x for x in result if x not in unneeded_results]

    def __init__(self, *args, **kwargs):
        self.party_path = get_party_path()
        self._depends_on = []
        self._sources = []
        self._source = None
        self._cache_manifest = {}
        self._last_parsed_source = None
        self._sources_manifest = {}
        self._last_unpacked_archive_info = {}
        self._builds = set()
        self._builds_manifest = {}

        super(Attendee, self).__init__(*args, **kwargs)

    def __repr__(self):
        """
        Get a representation of the attendee.
        """

        return 'Attendee(%r)' % self.name

    def depends_on(self, *attendees):
        """
        Make the attendee depend on one or several other attendees.

        `attendees` is a list of attendees to depend on.
        """

        self._depends_on.extend(attendees)
        return self

    @property
    def parents(self):
        """
        Get attendees this attendee directly depends on.
        """

        try:
            return set(self.get_enabled_instances(keys_list=self._depends_on))
        except Attendee.NoSuchInstance as ex:
            raise TeapotError(
                (
                    "Reference to a non-existing parent attendee %s could not "
                    "be resolved in attendee %s."
                ),
                hl(ex.keys[0]),
                hl(self),
            )

    @property
    def children(self):
        """
        Get attendees that directly depends on this attendee.
        """

        return {
            attendee for attendee in self.get_enabled_instances()
            if self in attendee.parents
        }

    @property
    def cache_path(self):
        return from_user_path(os.path.join(get_option('cache_root'), self.name))

    @property
    def sources_path(self):
        return from_user_path(os.path.join(get_option('sources_root'), self.name))

    @property
    def builds_path(self):
        return from_user_path(os.path.join(get_option('builds_root'), self.name))

    @property
    def cache_manifest_path(self):
        return os.path.join(self.cache_path, 'manifest.json')

    @property
    def cache_last_parsed_source_path(self):
        return os.path.join(self.cache_path, 'last_parsed_source.json')

    @property
    def sources_manifest_path(self):
        return os.path.join(self.sources_path, 'manifest.json')

    @property
    def sources_last_unpacked_archive_info_path(self):
        return os.path.join(self.sources_path, 'last_unpacked_archive_info.json')

    @property
    def builds_manifest_path(self):
        return os.path.join(self.builds_path, 'manifest.json')

    @property
    def cache_manifest(self):
        if not self._cache_manifest:
            try:
                self._cache_manifest = json.load(open(self.cache_manifest_path))

            except IOError:
                pass

        if not isinstance(self._cache_manifest, dict):
            self._cache_manifest = {}

        return self._cache_manifest

    @cache_manifest.setter
    def cache_manifest(self, value):
        self._cache_manifest = value

        mkdir(self.cache_path)
        json.dump(self._cache_manifest, open(self.cache_manifest_path, 'w'))

    @property
    def last_parsed_source(self):
        if not self._last_parsed_source:
            try:
                self._last_parsed_source = json.load(open(self.cache_last_parsed_source_path))

            except IOError:
                pass

        return self._last_parsed_source

    @last_parsed_source.setter
    def last_parsed_source(self, value):
        self._last_parsed_source = value

        mkdir(self.cache_path)
        json.dump(self._last_parsed_source, open(self.cache_last_parsed_source_path, 'w'))

    @property
    def sources_manifest(self):
        if not self._sources_manifest:
            try:
                self._sources_manifest = json.load(open(self.sources_manifest_path))

            except IOError:
                pass

        if not isinstance(self._sources_manifest, dict):
            self._sources_manifest = {}

        return self._sources_manifest

    @sources_manifest.setter
    def sources_manifest(self, value):
        self._sources_manifest = value

        mkdir(self.sources_path)
        json.dump(self._sources_manifest, open(self.sources_manifest_path, 'w'))

    @property
    def last_unpacked_archive_info(self):
        if not self._last_unpacked_archive_info:
            try:
                self._last_unpacked_archive_info = json.load(open(self.sources_last_unpacked_archive_info_path))

            except IOError:
                pass

        return self._last_unpacked_archive_info

    @last_unpacked_archive_info.setter
    def last_unpacked_archive_info(self, value):
        self._last_unpacked_archive_info = value

        mkdir(self.sources_path)
        json.dump(self._last_unpacked_archive_info, open(self.sources_last_unpacked_archive_info_path, 'w'))

    @property
    def builds_manifest(self):
        if not self._builds_manifest:
            try:
                self._builds_manifest = json.load(open(self.builds_manifest_path))

            except IOError:
                pass

        if not isinstance(self._builds_manifest, dict):
            self._builds_manifest = {}

        return self._builds_manifest

    @builds_manifest.setter
    def builds_manifest(self, value):
        self._builds_manifest = value

        mkdir(self.builds_path)
        json.dump(self._builds_manifest, open(self.builds_manifest_path, 'w'))

    @property
    def archive_path(self):
        return self.cache_manifest.get('archive_path')

    @property
    def archive_type(self):
        return tuple(self.cache_manifest.get('archive_type', []))

    @property
    def extracted_sources_path(self):
        return self.sources_manifest.get('extracted_sources_path')

    @property
    def must_fetch(self):
        # If the best source changed since last build, `self.source`
        # will clean the cache and the next test will evaluate to true.
        if not self.source:
            return True

        return not self.archive_path or not os.path.isfile(self.archive_path)

    @property
    def must_unpack(self):
        if not self.check_archive_signature():
            return True

        return not self.extracted_sources_path or not os.path.isdir(self.extracted_sources_path)

    @property
    def must_build(self):
        for build in self.builds:
            if self.builds_manifest.get(build.name) != build.signature:
                return True

        return False

    @property
    def sources(self):
        """
        Get all the active sources.
        """

        return [x for x in self._sources if x.enabled]

    def add_source(self, resource, *args, **kwargs):
        """
        Add a source to the attendee.

        `resource` is the resource to add to the attendee.
        """

        if self._source is not None:
            LOGGER.debug(
                "Clearing best source %s because a new source was added (%s).",
                hl(self._source),
                hl(source),
            )

            self._source = None

        if not isinstance(resource, Source):
            resource = Source(self, resource, *args, **kwargs)
        else:
            self._sources.append(resource)

        return self

    def get_source(self, resource):
        """
        Get a source.
        """

        return Source.get_instance_or_fail(self, resource)

    @property
    def builds(self):
        """
        Get all the active builds.
        """

        return [x for x in self._builds if x.enabled]

    def add_build(self, build, *args, **kwargs):
        """
        Add a build to the attendee.
        """

        if not isinstance(build, Build):
            build = Build(self, build, *args, **kwargs)
        else:
            self._builds.add(build)

        return self

    def get_build(self, build):
        """
        Get a build.
        """

        return Build.get_instance_or_fail(self, build)

    def clean(self):
        """
        Clean everything.
        """

        self.clean_cache()
        self.clean_sources()
        self.clean_builds()

    def clean_cache(self):
        """
        Clean the cache.
        """

        if os.path.exists(self.cache_path):
            LOGGER.info("Cleaning cache directory for %s.", hl(self))
            LOGGER.debug(
                "Cache directory for %s is at %s.",
                hl(self),
                hl(self.cache_path),
            )

            rmdir(self.cache_path)
        else:
            LOGGER.debug(
                "Cache directory for %s does not exist at %s. Nothing to do.",
                hl(self),
                hl(self.cache_path),
            )
            LOGGER.info("Cache directory for %s is already cleaned.", hl(self))

    def clean_sources(self):
        """
        Clean the sources.
        """

        if os.path.exists(self.sources_path):
            LOGGER.info("Cleaning sources directory for %s.", hl(self))
            LOGGER.debug(
                "Sources directory for %s is at %s.",
                hl(self),
                hl(self.sources_path),
            )

            rmdir(self.sources_path)
        else:
            LOGGER.debug(
                "Sources directory for %s does not exist at %s. Nothing to do.",
                hl(self),
                hl(self.sources_path),
            )
            LOGGER.info("Sources directory for %s is already cleaned.", hl(self))

    def clean_builds(self):
        """
        Clean the builds.
        """

        if os.path.exists(self.builds_path):
            LOGGER.info("Cleaning builds directory for %s.", hl(self))
            LOGGER.debug(
                "Builds directory for %s is at %s.",
                hl(self),
                hl(self.builds_path),
            )

            rmdir(self.builds_path)
        else:
            LOGGER.debug(
                "Builds directory for %s does not exist at %s. Nothing to do.",
                hl(self),
                hl(self.builds_path),
            )
            LOGGER.info("Builds directory for %s is already cleaned.", hl(self))

    @property
    def source(self):
        """
        Get the best source for the attendee.
        """

        if self._source is None:
            LOGGER.debug(
                "Finding best source for %s...",
                hl(self),
            )

            if not self.sources:
                raise TeapotError(
                    (
                        "No enabled source was found for the attendee %s. "
                        "Did you forget to set a filter on the attendee ?"
                    ),
                    hl(self),
                )

            for source in self.sources:
                try:
                    parsed_source = source.parsed_source

                    if parsed_source:
                        if self.last_parsed_source != parsed_source:
                            if self.last_parsed_source:
                                LOGGER.info(
                                    (
                                        "Best source has changed for %s: cleaning "
                                        "cache to trigger a refetch."
                                    ),
                                    hl(self),
                                )
                            else:
                                LOGGER.info(
                                    (
                                        "No previous best source found for %s: "
                                        "making sure the cache is cleaned so a "
                                        "fetch will occur."
                                    ),
                                    hl(self),
                                )

                            self.clean_cache()
                            self.last_parsed_source = parsed_source

                        LOGGER.debug(
                            "Best source for %s is now %s.",
                            hl(self),
                            hl(source),
                        )

                        self._source = source
                        break

                except TeapotError as ex:
                    LOGGER.warning("Error parsing source %s: " + ex.msg, hl(source), *ex.args)
                except Exception as ex:
                    LOGGER.warning("Error parsing source %s: %s", hl(source), hl(str(ex)))

        return self._source

    def check_archive_signature(self):
        """
        Check the archive signature.
        """

        m = hashlib.sha1()
        with open(self.archive_path) as f:
            for s in iter(lambda: f.read(1024 ** 2), ''):
                m.update(s)

        archive_signature = m.hexdigest()

        def update_signature():
            # This has to be done in this order or it won't work.
            self.clean_sources()

            LOGGER.debug(
                "Writing new archive signature for %s: %s",
                hl(self.archive_path),
                hl(archive_signature),
            )

            self.last_unpacked_archive_info = {
                'archive_signature': archive_signature,
            }

        archive_info = self.last_unpacked_archive_info

        if not archive_info:
            LOGGER.info(
                (
                    "No previous archive signature found for %s. Making sure "
                    "the sources directory gets cleaned."
                ),
                hl(self),
            )
            update_signature()
            return False

        last_archive_signature = archive_info.get('archive_signature')

        if archive_signature != last_archive_signature:
            LOGGER.info(
                (
                    "Archive signature for %s changed from %s to %s. Cleaning "
                    "the sources to make sure they get unpacked again."
                ),
                hl(self),
                hl(last_archive_signature),
                hl(archive_signature),
            )

            update_signature()
            return False

        LOGGER.debug(
            (
                "Archive signature for %s matches the last known one (%s). No "
                "cleaning of the sources needed."
            ),
            hl(self.archive_path),
            hl(archive_signature),
        )

        return True

    def fetch(self, force=False):
        """
        Fetch the most appropriate source.

        If `force` is truthy, the archive will be force-fetched again.
        """

        source = self.source

        if self.archive_path:
            if os.path.isfile(self.archive_path):
                if force:
                    LOGGER.info("%s was already fetched but force fetching was requested. Will fetch it again.", hl(self))
                    self.cache_manifest = {}
                else:
                    LOGGER.info("%s was already fetched. Nothing to do.", hl(self))
            else:
                LOGGER.warning("%s was already fetched according to its download manifest but the archive file (%s) was not found. Will fetch it again.", hl(self), self.archive_path)
                self.cache_manifest = {}
        else:
            LOGGER.info("No download manifest found for %s. Will fetch it.", hl(self))

        if not self.cache_manifest:
            LOGGER.info('Fetching %s from %s...', hl(self), hl(source))

            mkdir(self.cache_path)

            self.cache_manifest = source.fetch(target_path=self.cache_path)

            LOGGER.debug("Wrote new cache manifest for %s at: %s", hl(self), hl(self.cache_manifest_path))
            LOGGER.info("%s fetched successfully.", hl(self))

        if not self.cache_manifest:
            raise TeapotError(
                (
                    "All sources failed for the attendee %s. You may want "
                    "to check your network connectivity."
                ),
                hl(self),
            )

        LOGGER.debug(
            "Archive for %s (%s) is at %s",
            hl(self),
            hl(Unpacker.mimetype_to_str(self.archive_type)),
            hl(self.archive_path),
        )

    def unpack(self, force=False):
        """
        Unpack the archive.

        If `force` is truthy, the archive will be force-unpacked again.
        """

        if self.extracted_sources_path:
            if os.path.isdir(self.extracted_sources_path):
                if force:
                    LOGGER.info("%s was already unpacked but force unpacking was requested. Will unpack it again.", hl(self))
                    self.sources_manifest = {}
                else:
                    LOGGER.info("%s was already unpacked. Nothing to do.", hl(self))
            else:
                LOGGER.warning("%s was already unpacked according to its sources manifest but the extracted sources folder (%s) was not found. Will unpack it again.", hl(self), self.extracted_sources_path)
                self.sources_manifest = {}
        else:
            LOGGER.info("No sources manifest found for %s. Will fetch it.", hl(self))

        if not self.sources_manifest:
            LOGGER.info('Unpacking %s...', hl(self))

            mkdir(self.sources_path)

            LOGGER.debug(
                "Searching appropriate unpacker for archive %s of type %s...",
                hl(self.archive_path),
                hl(Unpacker.mimetype_to_str(self.archive_type)),
            )

            unpacker = Unpacker.get_instance_or_fail(self.archive_type)
            self.sources_manifest = unpacker.unpack(archive_path=self.archive_path, target_path=self.sources_path)

            LOGGER.debug('Clearing builds manifest as unpacking just took place.')
            self.builds_manifest = {}

        LOGGER.debug(
            "Archive for %s (%s) is unpacked at %s.",
            hl(self),
            hl(self.archive_path),
            hl(self.extracted_sources_path),
        )

    def build(self, force=False, verbose=False, keep_builds=False):
        """
        Build the attendee.

        If `force` is truthy, the archive will be force-built again.
        """

        for build in self.builds:
            last_signature = self.builds_manifest.get(build.name)
            signature = build.signature

            if last_signature is None:
                LOGGER.info(
                    "No known build signature for %s. Will build it.",
                    hl(build),
                )
            elif last_signature != signature:
                LOGGER.info(
                    "Last signature for %s (%s) does not match the current one (%s). Will build it again.",
                    hl(build),
                    hl(last_signature),
                    hl(signature),
                )
            elif force:
                LOGGER.info(
                    "Last signature for %s matches the current one (%s) but force-build requested. Will build it again.",
                    hl(build),
                    hl(signature),
                )
                self.builds_manifest[build.name] = None
            else:
                LOGGER.info(
                    "%s was built already. Nothing to do.",
                    hl(build),
                )
                return

            build_path = os.path.join(self.builds_path, build.name)
            log_path = os.path.join(self.builds_path, build.name + '.log')

            with temporary_copy(self.extracted_sources_path, build_path, persistent=keep_builds):
                build.build(path=build_path, log_path=log_path, verbose=verbose)
                self.builds_manifest[build.name] = signature



"""
Defines mechanism to handle options.
"""

from .filters import FilteredObject
from .filters import f
from .error import TeapotError
from .log import LOGGER
from .log import Highlight as hl


_OPTIONS = {}


class Option(object):
    """
    Defines an option.
    """

    class Value(FilteredObject):
        """
        An option value class.
        """

        def __init__(self, value, *args, **kwargs):
            super(Option.Value, self).__init__(*args, **kwargs)
            self.value = value

    @classmethod
    def get(cls, name):
        option = _OPTIONS.get(name)

        if not option:
            raise TeapotError(
                (
                    "No such registered option \"%s\". Did you forget to "
                    "call `register_option()` ?"
                ),
                name,
            )

        return option

    def __init__(self, name, value_type=str, default_values=[]):
        self.name = name
        self.value_type = value_type
        self._default_values = default_values
        self._values = []
        self._cached_value = None

    @property
    def value(self):
        """
        Get the option value.
        """

        if self._cached_value is None:
            values = Option.Value.get_enabled_instances(instances=self._values)

            if not values:
                values = Option.Value.get_enabled_instances(instances=self._default_values)

                if not values:
                    LOGGER.debug(
                        (
                            "Value request for option %s but none was found "
                            "that matches the current filters. Returning %s "
                            "instead."
                        ),
                        hl(self.name),
                        hl(repr(None)),
                    )
                else:
                    LOGGER.debug(
                        (
                            "Value requested for option %s but none was found "
                            "that matches the current filters. Using defaults "
                            "instead."
                        ),
                        hl(self.name),
                    )

            if len(values) > 1:
                raise TeapotError(
                    (
                        "More than one value matches the current filters "
                        "for option %s. Did you forget to set a filter on "
                        "some options ?"
                    ),
                    hl(self.name),
                )
            elif values:
                self._cached_value = values[0].value

                LOGGER.debug(
                    "Caching value %s for option %s.",
                    hl(self._cached_value),
                    hl(self.name),
                )

        return self._cached_value

    def add_value(self, value, *args, **kwargs):
        """
        Add a value.

        `value` is the value to add.

        Values are filterable, so you can also specify a `filter` for
        each added value.
        """

        self._values.append(Option.Value(self.value_type(value), *args, **kwargs))

        if self._cached_value is not None:
            LOGGER.debug(
                "Clearing cached value %s for option %s as its values just changed.",
                hl(self._cached_value),
                hl(self.name),
            )

            self._cached_value = None


def register_option(name, *args, **kwargs):
    """
    Register an option.

    `name` is the name of the option to register. If an option with that
    name already exists, an exception is thrown.
    """

    if name in _OPTIONS:
        raise TeapotError(
            "The option %s was already registered.",
            hl(name),
        )

    _OPTIONS[name] = Option(name, *args, **kwargs)


def set_option(name, value, *args, **kwargs):
    """
    Set an option.

    `name` is the name of the option to set.
    `value` is the value to set.

    You may specify a `filter` for each value.
    """

    Option.get(name).add_value(value, *args, **kwargs)


def get_option(name):
    """
    Get the value of the option with the specified `name`.
    """

    return Option.get(name).value

# Some built-in options.

register_option('cache_root', default_values=[
    Option.Value('~/.teapot/cache', filter=~f('windows')),
    Option.Value('%APPDATA%\\teapot\\cache', filter=f('windows')),
])
register_option('sources_root', default_values=[
    Option.Value('~/.teapot/sources', filter=~f('windows')),
    Option.Value('%APPDATA%\\teapot\\sources', filter=f('windows')),
])
register_option('builds_root', default_values=[
    Option.Value('~/.teapot/builds', filter=~f('windows')),
    Option.Value('%APPDATA%\\teapot\\builds', filter=f('windows')),
])
register_option('prefix', default_values=[
    Option.Value('~/.teapot/install', filter=~f('windows')),
    Option.Value('%APPDATA%\\teapot\\install', filter=f('windows')),
])



"""
Contains all teapot fetchers logic.
"""

from teapot.fetchers.fetcher import Fetcher
from teapot.fetchers.fetcher import register_fetcher
from teapot.fetchers.file_fetcher import FileFetcher  # NOQA
from teapot.fetchers.http_fetcher import HttpFetcher  # NOQA
from teapot.fetchers.github_fetcher import GithubFetcher  # NOQA



"""
A build class.
"""

import os
import re
import signal
import subprocess
import math

from datetime import datetime
from contextlib import contextmanager
from threading import Thread
from datetime import datetime

from .memoized import MemoizedObject
from .error import TeapotError
from .log import LOGGER, Highlight as hl
from .log import print_normal, print_error
from .filters import FilteredObject
from .environment import Environment
from .extensions import parse_extension
from .prefix import PrefixedObject
from .signature import SignableObject


class Build(MemoizedObject, FilteredObject, PrefixedObject, SignableObject):

    """
    Represents a build.
    """

    memoization_keys = ('attendee', 'name')
    propagate_memoization_keys = True
    signature_fields = ('environment', 'subdir', 'commands', 'filter')

    @classmethod
    def transform_memoization_keys(cls, attendee, name):
        """
        Make sure the attendee parameter is a real Attendee instance.
        """

        if isinstance(attendee, basestring):
            from .attendee import Attendee

            attendee = Attendee(attendee)

        return attendee, name

    def __init__(self, attendee, name, environment, subdir=None, commands=None, *args, **kwargs):
        super(Build, self).__init__(*args, **kwargs)
        self._environment = environment
        self.subdir = subdir
        self.commands = commands or []

        # Register the build in the Attendee.
        self.attendee = attendee
        self.name = name
        attendee.add_build(self)

    @property
    def environment(self):
        if not isinstance(self._environment, Environment):
            self._environment = Environment.get_instance(self._environment)

        return self._environment

    def __repr__(self):
        """
        Get a representation of the build.
        """

        return 'Build(%r, %r)' % (self.attendee, self.name)

    def __str__(self):
        """
        Get a string representation of the build.
        """

        return '%s_%s' % (self.attendee, self.name)

    def add_command(self, command):
        """
        Add a command to the builder.
        """

        self.commands.append(command)

    @contextmanager
    def create_log_file(self, log_path):
        """
        Create a log file object and returns it.

        `log_path` is the path to the log file to write to.
        """

        try:
            LOGGER.debug('Opening log file at: %s', hl(log_path))

            with open(log_path, 'w') as log_file:
                yield log_file

        finally:
            LOGGER.info('Log file written to: %s', hl(log_path))

    @contextmanager
    def chdir(self, path):
        """
        Change the current directory.
        """

        old_path = os.getcwd()

        LOGGER.debug('Moving to: %s', hl(path))
        os.chdir(path)

        try:
            yield path
        finally:
            LOGGER.debug('Moving back to: %s', hl(old_path))
            os.chdir(old_path)

    @contextmanager
    def handle_interruptions(self, callable=None):
        """
        Handle interruptions.
        """

        def handler(signum, frame):
            LOGGER.warning('The building process was interrupted by the user.')

            if callable:
                callable()

        previous_handler = signal.signal(signal.SIGINT, handler)

        try:
            yield
        finally:
            signal.signal(signal.SIGINT, previous_handler)

    def build(self, path, log_path, verbose=False):
        """
        Launch the build in the specified `path`.

        `log_path` is the path to the log file to create.
        """

        working_dir = os.path.join(path, self.subdir if self.subdir else '')

        with self.create_log_file(log_path) as log_file:
            with self.chdir(working_dir):
                with self.environment.enable() as env:
                    LOGGER.info("Build started in %s at %s.", hl(working_dir), hl(datetime.now().strftime('%c')))
                    log_file.write("Build started in %s at %s.\n" % (working_dir, datetime.now().strftime('%c')))

                    if env.shell:
                        LOGGER.info('Building within: %s', hl(' '.join(env.shell)))
                        log_file.write('Using "%s" as a shell.\n' % ' '.join(env.shell))
                    else:
                        LOGGER.info('Building within %s.', hl('the default system shell'))
                        log_file.write('Using system shell.\n')

                    for key, value in os.environ.iteritems():
                        LOGGER.debug('%s: %s', key, hl(value))
                        log_file.write('%s: %s\n' % (key, value))

                    for index, command in enumerate(self.commands):
                        command = self.apply_extensions(command)
                        numbered_prefix = ('%%0%sd' % int(math.ceil(math.log10(len(self.commands))))) % index

                        LOGGER.important('%s: %s', numbered_prefix, hl(command))
                        log_file.write('%s: %s\n' % (numbered_prefix, command))

                        if env.shell:
                            process = subprocess.Popen(env.shell + [command], shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        else:
                            process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

                        mixed_output = []

                        with self.handle_interruptions(process.terminate):
                            def read_stdout():
                                for line in iter(process.stdout.readline, ''):
                                    mixed_output.append((print_normal, line))
                                    log_file.write(line)

                                    if verbose:
                                        print_normal(line)

                            def read_stderr():
                                for line in iter(process.stderr.readline, ''):
                                    mixed_output.append((print_error, line))
                                    log_file.write(line)

                                    if verbose:
                                        print_error(line)

                            stdout_thread = Thread(target=read_stdout)
                            stdout_thread.daemon = True
                            stdout_thread.start()

                            stderr_thread = Thread(target=read_stderr)
                            stderr_thread.daemon = True
                            stderr_thread.start()

                            map(Thread.join, [stdout_thread, stderr_thread])
                            process.wait()

                        log_file.write('\n')

                        if process.returncode != 0:
                            if not verbose:
                                for func, line in mixed_output:
                                    func(line)

                            log_file.write('Command failed with status: %s\n' % process.returncode)
                            log_file.write('Build failed at %s.\n' % datetime.now().strftime('%c'))

                            raise subprocess.CalledProcessError(returncode=process.returncode, cmd=command)

                    LOGGER.info("Build succeeded at %s.", hl(datetime.now().strftime('%c')))
                    log_file.write("Build succeeded at %s.\n" % datetime.now().strftime('%c'))

    def apply_extensions(self, command):
        """
        Apply the extensions to the command.
        """

        def replace(match):
            extension, args, kwargs = parse_extension(match.group('extension'))
            return extension(self, *args, **kwargs)

        return re.sub(r'\{{(?P<extension>([^}]|[}][^}])+)}}', replace, command)



"""
The teapot main script.
"""

import os
import sys
import argparse
import logging
import traceback

from functools import wraps

from .log import LOGGER, ColorizingStreamHandler
from .error import TeapotError
from .path import chdir

import teapot.party


def main():
    """
    The script entry point.
    """

    parser = argparse.ArgumentParser(
        description='Manage third-party software.'
    )

    parser.add_argument('-d', '--debug', action='store_true',
                        help='Enable debug output.')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Be more explicit about what happens.')

    command_parser = parser.add_subparsers(help='The available commands.')
    parser.add_argument('-p', '--party-file', default=None,
                        help='The party-file to read.')

    # The clean command
    clean_command_parser = command_parser.add_parser(
        'clean', help='Clean the party.')
    clean_subcommand_parser = clean_command_parser.add_subparsers(
        help='The available commands.')

    #  The clean cache subcommand
    clean_cache_command_parser = clean_subcommand_parser.add_parser(
        'cache', help='Clean the cache.')
    clean_cache_command_parser.set_defaults(func=clean_cache)
    clean_cache_command_parser.add_argument(
        'attendees', metavar='attendee', nargs='*', default=[], help='The attendees to clean.')

    #  The clean sources subcommand
    clean_sources_command_parser = clean_subcommand_parser.add_parser(
        'sources', help='Clean the sources.')
    clean_sources_command_parser.set_defaults(func=clean_sources)
    clean_sources_command_parser.add_argument(
        'attendees', metavar='attendee', nargs='*', default=[], help='The attendees to clean.')

    #  The clean builds subcommand
    clean_builds_command_parser = clean_subcommand_parser.add_parser(
        'builds', help='Clean the builds.')
    clean_builds_command_parser.set_defaults(func=clean_builds)
    clean_builds_command_parser.add_argument(
        'attendees', metavar='attendee', nargs='*', default=[], help='The attendees to clean.')

    #  The clean all subcommand
    clean_all_command_parser = clean_subcommand_parser.add_parser(
        'all', help='Clean the party cache, build and install.')
    clean_all_command_parser.set_defaults(func=clean_all)
    clean_all_command_parser.add_argument(
        'attendees', metavar='attendee', nargs='*', default=[], help='The attendees to clean.')

    # The fetch command
    fetch_command_parser = command_parser.add_parser(
        'fetch', help='Fetch all the archives.')
    fetch_command_parser.set_defaults(func=fetch)
    fetch_command_parser.add_argument(
        'attendees', metavar='attendee', nargs='*', default=[], help='The attendees to fetch.')
    fetch_command_parser.add_argument(
        '-f', '--force', action='store_true', help='Fetch archives even if they already exist in the cache.')

    # The unpack command
    unpack_command_parser = command_parser.add_parser(
        'unpack', help='Unpack all the fetched archives.')
    unpack_command_parser.set_defaults(func=unpack)
    unpack_command_parser.add_argument(
        'attendees', metavar='attendee', nargs='*', default=[], help='The attendees to unpack.')
    unpack_command_parser.add_argument(
        '-f', '--force', action='store_true', help='Unpack archives even if they already exist in the build.')

    # The build command
    build_command_parser = command_parser.add_parser(
        'build', help='Build the archives.')
    build_command_parser.set_defaults(func=build)
    build_command_parser.add_argument(
        'attendees', metavar='attendee', nargs='*', default=[], help='The attendees to build.')
    build_command_parser.add_argument(
        '-f', '--force', action='store_true', help='Build archives even if they were already built.')
    build_command_parser.add_argument(
        '-k', '--keep-builds', action='store_true', help='Keep the build directories for inspection.')

    args = parser.parse_args()

    handler = ColorizingStreamHandler(sys.stdout)
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)

    if args.debug:
        logging.getLogger().addHandler(handler)
        logging.getLogger().setLevel(logging.DEBUG)
        logging.getLogger('lepl').setLevel(logging.WARNING)
    else:
        LOGGER.addHandler(handler)
        LOGGER.setLevel(logging.INFO)

    if args.party_file is None:
        party_files = ('Party', 'Party.py', '.party', '.party.py')
        existing_files = filter(os.path.isfile, party_files)

        if not existing_files:
            LOGGER.error(
                'No party-file was found. (Looked for: %s)',
                ', '.join(map(repr, party_files)),
            )

            return 1

        args.party_file = os.path.join(os.getcwd(), existing_files[0])
    else:
        args.party_file = os.path.abspath(args.party_file)

    with chdir(os.path.dirname(args.party_file)):
        try:
            teapot.party.load_party_file(args.party_file)

        except IOError:
            LOGGER.error(
                'No party-file was found. (Searched path: "%s")',
                args.party_file,
            )

            return 1

        if not args.func(args):
            return 2


def command(func):
    """
    Provides exception handling for commands.
    """

    @wraps(func)
    def decorated(args, **kwargs):
        try:
            func(args, **kwargs)

            return True

        except KeyboardInterrupt:
            pass

        except TeapotError as ex:
            LOGGER.error(ex.msg, *ex.msg_args)

            if args.debug:
                LOGGER.debug('\nTraceback is:\n' + ''.join(traceback.format_tb(sys.exc_info()[2])))

        except Exception as ex:
            if args.debug:
                LOGGER.exception(ex)
            else:
                LOGGER.error(str(ex))

        LOGGER.error('teapot execution failed.')

        return False

    return decorated


@command
def clean_cache(args):
    """
    Clean the cache.
    """

    teapot.party.clean_cache(
        attendees=args.attendees,
    )


@command
def clean_sources(args):
    """
    Clean the sources.
    """

    teapot.party.clean_sources(
        attendees=args.attendees,
    )


@command
def clean_builds(args):
    """
    Clean the builds.
    """

    teapot.party.clean_builds(
        attendees=args.attendees,
    )


@command
def clean_all(args):
    """
    Clean the party cache and build.
    """

    teapot.party.clean_all(
        attendees=args.attendees,
    )


@command
def fetch(args):
    """
    Fetch the archives.
    """

    teapot.party.fetch(
        attendees=args.attendees,
        force=args.force,
    )


@command
def unpack(args):
    """
    Unpack the archives.
    """

    teapot.party.unpack(
        attendees=args.attendees,
        force=args.force,
    )


@command
def build(args):
    """
    Build the archives.
    """

    teapot.party.build(
        attendees=args.attendees,
        force=args.force,
        verbose=args.verbose,
        keep_builds=args.keep_builds,
    )



"""
Contains all teapot unpackers logic.
"""

from teapot.unpackers.unpacker import Unpacker
from teapot.unpackers.unpacker import register_unpacker
from teapot.unpackers.tarball_unpacker import TarballUnpacker
from teapot.unpackers.zipfile_unpacker import ZipFileUnpacker



"""
A teapot error class.
"""


class TeapotError(RuntimeError):

    """
    Represents a runtime teapot-specific error.
    """

    def __init__(self, msg, *args):
        self.msg = msg
        self.msg_args = args

    def __str__(self):
        return self.msg % self.msg_args



"""
Some global variables.
"""

import os

from contextlib import contextmanager


GLOBALS = {
    'PARTY_PATH': None,
}


@contextmanager
def set_party_path(path):
    old_party_path, GLOBALS['PARTY_PATH'] = GLOBALS['PARTY_PATH'], os.path.abspath(path)

    try:
        yield
    finally:
        GLOBALS['PARTY_PATH'] = old_party_path


def get_party_path():
    return GLOBALS['PARTY_PATH']



"""
A base filter class.
"""

from functools import wraps

from ..memoized import MemoizedObject
from ..error import TeapotError
from ..log import Highlight as hl


class UnamedFilter(object):
    """
    A generic filter class.
    """

    def __init__(self, condition=None):
        self._condition = condition

    def __nonzero__(self):
        if self._condition:
            try:
                return bool(self._condition())
            except TypeError:
                if isinstance(self._condition, basestring):
                    return bool(Filter.get_instance_or_fail(self._condition))

                return bool(self._condition)

        return False

    def __and__(self, other):
        return UnamedFilter(condition=lambda: self and other)

    def __or__(self, other):
        return UnamedFilter(condition=lambda: self or other)

    def __invert__(self):
        return UnamedFilter(condition=lambda: not self)

    def __xor__(self, other):
        return UnamedFilter(condition=lambda: (self or other) and not (self and other))


class Filter(MemoizedObject, UnamedFilter):

    propagate_memoization_keys = True
    no_such_instance_message = "No filter named %s could be found. Did you mistype the filter's name ?"
    no_such_instance_args = ('key_name',)

    def __init__(self, name, condition=None):
        if condition is None:
            raise TeapotError(
                (
                    "A filter definition for %s must contain a non-None "
                    "condition. Did you mistype the filter's name ?"
                ),
                hl(name),
            )

        super(Filter, self).__init__(condition=condition)


class named_filter(object):

    """
    A filter-maker decorator.
    """

    def __init__(self, name, depends_on=None):
        """
        Define a named filter from a function.

        `name` is the name of the filter to create. If a filter with
        that name already exists, a TeapotError is raised.

        `depends_on` is another filter to depend on. It can be either a
        filter instance or the name of a named filter.
        """

        self.name = name
        self.depends_on = depends_on

    def __call__(self, func):
        if self.depends_on:
            @wraps(func)
            def depends_on_condition():
                if isinstance(self.depends_on, basestring):
                    depends_on_filter = Filter.get_instance_or_fail(self.depends_on)

                    if not depends_on_filter:
                        return False
                else:
                    if not self.depends_on:
                        return False

                return func()

            condition = depends_on_condition
        else:
            condition = func

        with Filter.raise_on_duplicate():
            Filter(name=self.name, condition=condition)

        return func


uf = UnamedFilter
f = Filter



"""
A teapot path-handling class.
"""

import os
import stat
import shutil
import errno

from contextlib import contextmanager
from functools import wraps

from teapot.log import LOGGER
from teapot.log import Highlight as hl


def from_user_path(path):
    """
    Perform all variables substitutions from the specified user path.
    """

    return os.path.normpath(os.path.expanduser(os.path.expandvars(path)))


def resolve_user_path(func):
    """
    A decorator that resolves user paths in the return value.
    """

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        return from_user_path(func(*args, **kwargs))

    return wrapped_func


def read_path(value, base_path, default_path):
    """
    Read a path value from a string.

    If `value` is a string, the Cache is default created at the location
    specified by `value`.

    If `value` is falsy, the Cache is default created.
    """

    if not value:
        cache_path = from_user_path(default_path)
    else:
        cache_path = from_user_path(value)

    if not os.path.isabs(cache_path):
        cache_path = os.path.normpath(os.path.join(base_path, cache_path))

    return cache_path


@contextmanager
def chdir(path):
    """
    Changes the directory temporarily.
    """

    path = os.path.abspath(path)
    saved_dir = os.getcwd()

    if os.path.abspath(saved_dir) != os.path.abspath(path):
        LOGGER.debug(
            "Temporarily changing current directory from %s to %s",
            hl(saved_dir),
            hl(path),
        )

        os.chdir(path)

    try:
        yield
    finally:
        if os.path.abspath(saved_dir) != os.path.abspath(path):
            LOGGER.debug(
                "Changing back current directory from %s to %s",
                hl(path),
                hl(saved_dir),
            )

            os.chdir(saved_dir)


def mkdir(path):
    """
    Create the specified path.

    Does nothing if the path exists.
    """

    try:
        if not os.path.isdir(path):
            LOGGER.debug('Creating directory at %s.', hl(path))

        os.makedirs(path)

    except OSError as ex:
        if ex.errno != errno.EEXIST or not os.path.isdir(path):
            raise


def rmdir(path):
    """
    Delete the specified path if it exists.

    Does nothing if the path doesn't exist.
    """

    try:
        LOGGER.info('Removing directory at %s.', hl(path))

        def onerror(func, path, excinfo):
            if os.path.exists(path):
                LOGGER.debug('Was unable to delete "%s": %s', hl(path), excinfo[1])
                LOGGER.debug('Trying again after changing permissions...')
                os.chmod(path, stat.S_IWUSR)

                try:
                    func(path)
                except Exception as ex:
                    LOGGER.error('Unable to delete "%s": %s', hl(path), excinfo[1])

                    raise

        shutil.rmtree(path, ignore_errors=False, onerror=onerror)

    except Exception as ex:
        LOGGER.warning(ex)


def copytree(src, dst, symlinks=False, ignore=None, copy_function=shutil.copy2):
    names = os.listdir(src)
    if ignore is not None:
        ignored_names = ignore(src, names)
    else:
        ignored_names = set()

    os.makedirs(dst)
    errors = []
    for name in names:
        if name in ignored_names:
            continue
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if symlinks and os.path.islink(srcname):
                linkto = os.readlink(srcname)
                os.symlink(linkto, dstname)
            elif os.path.isdir(srcname):
                copytree(srcname, dstname, symlinks, ignore, copy_function)
            else:
                try:
                    copy_function(srcname, dstname)
                except (IOError, WindowsError):
                    shutil.copy2(srcname, dstname)

            # XXX What about devices, sockets etc.?
        except (IOError, os.error) as why:
            errors.append((srcname, dstname, str(why)))
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except Error as err:
            errors.extend(err.args[0])
    try:
        shutil.copystat(src, dst)
    except WindowsError:
        # can't copy file access times on Windows
        pass
    except OSError as why:
        errors.extend((src, dst, str(why)))
    if errors:
        raise Error(errors)


@contextmanager
def temporary_copy(source_path, target_path, persistent=False):
    """
    Copy a source path to a target path.

    The target will be deleted upon function exist, unless `persistent`
    is truthy.
    """

    try:
        if os.path.exists(target_path):
            rmdir(target_path)

        LOGGER.info('Copying %s to %s...', hl(source_path), hl(target_path))
        copytree(source_path, target_path, copy_function=getattr(os, 'link', shutil.copy2))

        yield target_path

    finally:
        if not persistent:
            rmdir(target_path)
        else:
            LOGGER.info('Not erasing temporary directory at %s.', hl(target_path))


def windows_to_unix_path(path):
    """
    Convert a Windows path to a UNIX path, in such a way that it can be used in
    MSys or Cygwin.
    """

    drive, tail = os.path.splitdrive(path)

    if drive:
        drive = '/' + drive[0]

    return drive + tail.replace('\\', '/')



"""
A local-file fetcher class.
"""

import os
import shutil
import urlparse
import mimetypes

from teapot.fetchers.fetcher import register_fetcher
from teapot.fetchers.fetcher import FetcherImplementation
from teapot.path import rmdir


@register_fetcher('file')
class FileFetcher(FetcherImplementation):

    """
    Fetchs a file on the local filesystem.
    """

    def parse_source(self, source):
        """
        Checks that the `source` is a local filename.
        """

        parse = urlparse.urlsplit(source.resource)

        if parse.scheme == 'file':
            return {
                'file_path': os.path.abspath(os.path.join(parse.netloc, parse.path)),
            }

    def fetch(self, fetch_info, target_path, progress):
        """
        Fetch a file.
        """

        archive_path = os.path.join(target_path, os.path.basename(fetch_info['file_path']))

        size = os.path.getsize(fetch_info['file_path'])

        if not size:
            size = None

        rmdir(archive_path)

        progress.on_start(target=os.path.basename(archive_path), size=size)

        if os.path.isfile(fetch_info['file_path']):
            shutil.copyfile(fetch_info['file_path'], archive_path)
            archive_type = mimetypes.guess_type(fetch_info['file_path'], strict=False)
        elif os.path.isdir(fetch_info['file_path']):
            shutil.copytree(fetch_info['file_path'], archive_path)
            archive_type = (None, None)

        # No real interactive progress to show here.
        #
        # This could be fixed though.

        progress.on_update(progress=size)
        progress.on_finish()

        return {
            'archive_path': archive_path,
            'archive_type': archive_type,
        }



"""
An environment class.
"""

import os
import re
import sys

from contextlib import contextmanager

from .memoized import MemoizedObject
from .error import TeapotError
from .log import LOGGER, Highlight as hl
from .signature import SignableObject


class Environment(MemoizedObject, SignableObject):

    """
    Represents a build environment.
    """

    propagate_memoization_keys = True
    signature_fields = ('variables', 'shell', 'parent')

    @staticmethod
    def perform_substitutions(value, parent_context):
        """
        Perform substitutions in the specified `value`.

        Values are taken from the `parent_context`. If no value is found in the
        parent context, the substitution is done with an empty string.

        Substitutions are done for $KEY on all platforms, and also for %KEY% on
        Windows.
        """

        if sys.platform.startswith('win32'):
            pattern = '\$(?P<unix_key>[A-Za-z_][A-Za-z_0-9]*)|%(?P<windows_key>[A-Za-z_][A-Za-z_0-9]*)%'
        else:
            pattern = '\$(?P<unix_key>[A-Za-z_][A-Za-z_0-9]*)'

        def substitute(match):
            key = match.group('unix_key') or match.group('windows_key')

            return parent_context.get(key, '')

        return re.sub(pattern, substitute, value)

    def __init__(self, name, variables=None, shell=None, parent=None, *args, **kwargs):
        """
        Create a new environment.

        `variables` is a dictionary of environment variables to use.

        `shell`, if specified, is the path to a shell to use, or parts of a
        shell to use. Variables in `shell` will be substitued using the parent
        environment, if any. If `shell` is True, then the shell from the parent
        environment will be used.

        `parent` is an Environment instance or the name of an Environment
        instance to inherit from. If `parent` is None, the environment will
        be blank.
        """

        self._parent = parent
        self.variables = variables or {}
        self._shell = shell

        super(Environment, self).__init__(*args, **kwargs)

    def __repr__(self):
        """
        Get a representation of the environment.
        """

        return 'Environment(%r)' % self.name

    def depends_on(self, environment):
        """
        Make the environment depend on one other environment.

        `environment` is either a string or an Environment instance.
        """

        self._parent = environment
        return self

    @property
    def parent(self):
        """
        Get the parent environment, if any.
        """

        if isinstance(self._parent, basestring):
            self._parent = self.get_instance_or_fail(self._parent)
        elif isinstance(self._parent, Environment):
            pass
        elif self._parent is not None:
            raise TeapotError(
                "An invalid parent %s was specified for environment %s.",
                hl(repr(self._parent)),
                hl(self),
            )

        return self._parent

    @property
    def children(self):
        """
        Get environments that directly depend on this environment.
        """

        return {
            environment for environment in self.get_instances()
            if self == environment.parent
        }

    @property
    def shell(self):
        if self._shell is True:
            shell = self.parent._shell if self.parent else None
        elif isinstance(self._shell, basestring):
            shell = [self._shell]
        else:
            shell = self._shell

        if shell:
            if self.parent:
                with self.parent.enable():
                    return map(lambda arg: self.perform_substitutions(arg, os.environ), shell)
            else:
                return map(lambda arg: self.perform_substitutions(arg, {}), shell)

    @shell.setter
    def shell(self, value):
        self._shell = value

    @contextmanager
    def enable(self, silent=False):
        """
        Enable the environment and shell settings, if any.

        The current process environment is modified during the call, and
        restored when the function returns.

        If `silent` is truthy, no log output will be done.

        enable() is supposed to be used with a `with` statement.
        """

        saved_environ = os.environ.copy()

        try:
            if self.parent:
                with self.parent.enable(silent=True):
                    if not silent:
                        LOGGER.info('Entering environment %s...', hl(self))

                    for key, value in self.variables.iteritems():
                        if value is not None:
                            os.environ[key] = self.perform_substitutions(value, saved_environ)
                        else:
                            if key in os.environ:
                                del os.environ[key]

                    yield self
            else:
                if not silent:
                    LOGGER.info('Entering environment %s...', hl(self))

                os.environ.clear()

                for key, value in self.variables.iteritems():
                    if value is not None:
                        os.environ[key] = self.perform_substitutions(value, saved_environ)
                    else:
                        if key in saved_environ:
                            os.environ[key] = saved_environ.get(key)

                yield self

        finally:
            if not silent:
                LOGGER.info('Exiting environment %s...', hl(self))

            os.environ.clear()
            os.environ.update(saved_environ)


# Create the empty and system environments.
Environment('empty')
Environment('system', variables=os.environ.copy())



"""
A base Unpacker class.
"""

from ..memoized import MemoizedObject
from ..log import Highlight as hl

from .callbacks import ProgressBarUnpackerCallback


class UnpackerImplementation(object):

    """
    Base class for all unpacker implementation classes.
    """

    def unpack(self, archive_path, target_path, progress):
        """
        Unpack the specified archive.

        `archive_path` is the path of the archive to unpack.
        `target_path` is the path to extract the archive to.

        It returns the unpack manifest.
        """

        raise NotImplementedError


class Unpacker(MemoizedObject):

    """
    Represent an unpacker.
    """

    memoization_keys = ('mimetype',)
    no_such_instance_message = "Unable to find an unpacker that handles archives of type %s."
    no_such_instance_args = ('self',)

    @staticmethod
    def mimetype_to_str(mimetype):
        return '(%s)' % ','.join(x for x in mimetype if x)

    def __init__(self, unpacker_impl_class, progress_class=ProgressBarUnpackerCallback):
        """
        Create a new unpacker that uses the specified unpacker
        implementation.
        """

        self._unpacker_impl = unpacker_impl_class()
        self.progress = progress_class()

    def __str__(self):
        return mimetype_to_str(self.mimetype)

    def unpack(self, archive_path, target_path):
        """
        Unpack the specified archive.

        `archive_path` is the path of the archive to unpack.
        `target_path` is the path to extract the archive to.

        It returns the unpack manifest.
        """

        try:
            return self._unpacker_impl.unpack(
                archive_path=archive_path,
                target_path=target_path,
                progress=self.progress,
            )

        except Exception as ex:
            self.progress.on_exception(ex)

            raise


class register_unpacker(object):
    """
    A decorator that registers an unpacker.
    """

    def __init__(self, mimetype):
        self.mimetype = mimetype

    def __call__(self, cls):
        with Unpacker.raise_on_duplicate():
            Unpacker(mimetype=self.mimetype, unpacker_impl_class=cls)

        return cls



"""
The party classes.
"""

import os
import sys
import imp

from contextlib import contextmanager

from .log import LOGGER
from .log import Highlight as hl
from .attendee import Attendee
from .globals import set_party_path


@contextmanager
def disable_bytecode_generation():
    """
    Temporarily disable bytecode generation.
    """
    sentinel, sys.dont_write_bytecode = sys.dont_write_bytecode, True

    try:
        yield
    finally:
        sys.dont_write_bytecode = sentinel


def load_party_file(path):
    """
    Load a Party from a file.

    `path` must point to an existing party file.
    """

    path = os.path.abspath(path)

    LOGGER.debug('Importing party file from %r.', path)

    with disable_bytecode_generation():
        with set_party_path(path):
            imp.load_source('party', path)

    LOGGER.debug('Done importing party file.')


def clean_all(attendees=None):
    """
    Clean everything.
    """

    attendees = Attendee.get_enabled_instances(attendees or None)

    LOGGER.info("Cleaning everything for %s attendee(s)...", hl(len(attendees)))

    clean_cache(attendees)
    clean_sources(attendees)
    clean_builds(attendees)

    LOGGER.info("Done cleaning everything for %s attendee(s)...", hl(len(attendees)))


def clean_cache(attendees=None):
    """
    Clean the cache.
    """

    attendees = Attendee.get_enabled_instances(attendees or None)

    LOGGER.info("Cleaning cache for %s attendee(s)...", hl(len(attendees)))

    for attendee in attendees:
        attendee.clean_cache()

    LOGGER.info("Done cleaning cache for %s attendee(s)...", hl(len(attendees)))


def clean_sources(attendees=None):
    """
    Clean the sources.
    """

    attendees = Attendee.get_enabled_instances(attendees or None)

    LOGGER.info("Cleaning sources for %s attendee(s)...", hl(len(attendees)))

    for attendee in attendees:
        attendee.clean_sources()

    LOGGER.info("Done cleaning sources for %s attendee(s)...", hl(len(attendees)))


def clean_builds(attendees=None):
    """
    Clean the builds.
    """

    attendees = Attendee.get_enabled_instances(attendees or None)

    LOGGER.info("Cleaning builds for %s attendee(s)...", hl(len(attendees)))

    for attendee in attendees:
        attendee.clean_builds()

    LOGGER.info("Done cleaning builds for %s attendee(s)...", hl(len(attendees)))


def fetch(attendees=None, force=False):
    """
    Fetch the specified attendees.
    """

    attendees = Attendee.get_dependent_instances(attendees or None)

    if force:
        LOGGER.info("Force fetch requested...")
    else:
        attendees = [x for x in attendees if x.must_fetch]

        if not attendees:
            LOGGER.info("All attendees were fetched already. Nothing to do.")
            return

    LOGGER.info("Will now fetch %s." % ", ".join(["%s"] * len(attendees)), *map(hl, attendees))

    for attendee in attendees:
        attendee.fetch(force=force)

    LOGGER.info("Done fetching %s attendee(s)...", hl(len(attendees)))


def unpack(attendees=None, force=False):
    """
    Unpack the specified attendees.
    """

    fetch(attendees, force=force)

    attendees = Attendee.get_dependent_instances(attendees or None)

    if force:
        LOGGER.info("Force unpack requested...")
    else:
        attendees = [x for x in attendees if x.must_unpack]

        if not attendees:
            LOGGER.info("All attendees were unpacked already. Nothing to do.")
            return

    LOGGER.info("Will now unpack %s." % ", ".join(["%s"] * len(attendees)), *map(hl, attendees))

    for attendee in attendees:
        attendee.unpack(force=force)

    LOGGER.info("Done unpacking %s attendee(s)...", hl(len(attendees)))


def build(attendees=None, force=False, verbose=False, keep_builds=False):
    """
    Build the specified attendees.
    """

    unpack(attendees, force=force)

    attendees = Attendee.get_dependent_instances(attendees or None)

    if force:
        LOGGER.info("Force build requested...")
    else:
        attendees = [x for x in attendees if x.must_build]

        if not attendees:
            LOGGER.info("All attendees were built already. Nothing to do.")
            return

    LOGGER.info("Will now build %s." % ", ".join(["%s"] * len(attendees)), *map(hl, attendees))

    for attendee in attendees:
        attendee.build(force=force, verbose=verbose, keep_builds=keep_builds)

    LOGGER.info("Done building %s attendee(s)...", hl(len(attendees)))



"""
A Github fetcher class.
"""

import re

from teapot.fetchers.fetcher import register_fetcher
from teapot.fetchers.http_fetcher import HttpFetcher


@register_fetcher('github')
class GithubFetcher(HttpFetcher):

    """
    Fetchs an archive from Github.
    """

    def parse_source(self, source):
        """
        Checks that the `source` is a local filename.
        """

        match = re.match(r'github:(?P<owner>[\S/]+)/(?P<repo>[\S/]+)/(?P<ref>.+)', source.resource)

        if match:
            values = match.groupdict()
            values['archive_format'] = 'tarball'

            return {
                'url': 'https://api.github.com/repos/%(owner)s/%(repo)s/%(archive_format)s/%(ref)s' % values,
                'mimetype': source.mimetype,
            }



"""
Provides tools to deal with third-party software.
"""

import teapot.filters.builtin
import teapot.extensions.builtin

from .attendee import Attendee
from .filters import f
from .filters import uf
from .environment import Environment
from .options import register_option
from .options import set_option
from .options import get_option
from .build import Build
from .extensions.extension import register_extension



"""
An unpacker that deals with .tgz files.
"""

from teapot.unpackers.unpacker import UnpackerImplementation
from teapot.unpackers.unpacker import register_unpacker

from ..error import TeapotError
from ..log import Highlight as hl

import os
import tarfile


@register_unpacker(('application/x-gzip', None))
@register_unpacker(('application/x-bzip2', None))
class TarballUnpacker(UnpackerImplementation):

    """
    An unpacker class that deals with .tgz files.
    """

    def unpack(self, archive_path, target_path, progress):
        """
        Uncompress the archive.
        """

        if not tarfile.is_tarfile(archive_path):
            raise TeapotError(
                "%s is not a valid tar archive.",
                hl(archive_path),
            )

        tar = tarfile.open(archive_path, 'r')

        # We get the common prefix for all archive members.
        prefix = os.path.commonprefix(tar.getnames())

        # An archive member with the prefix as a name can exist in the archive or ends with a /.
        while not prefix.endswith('/'):
            try:
                prefix_member = tar.getmember(prefix)

                if prefix_member.isdir:
                    break

            except KeyError:
                pass

            new_prefix = os.path.dirname(prefix)

            if prefix == new_prefix:
                raise TeapotError(
                    "Unable to find a common prefix in %s to extract from.",
                    hl(archive_path),
                )
            else:
                prefix = new_prefix

        extracted_sources_path = os.path.join(target_path, prefix)

        progress.on_start(archive_path=archive_path, count=len(tar.getmembers()))

        for index, member in enumerate(tar.getmembers()):
            if os.path.isabs(member.name):
                raise TeapotError(
                    (
                        "Refusing to extract an archive (%s) that contains "
                        "absolute filenames."
                    ),
                    hl(archive_path),
                )

            progress.on_update(current_file=member.name, progress=index)
            tar.extract(member, path=target_path)

        progress.on_finish()

        return {
            'extracted_sources_path': extracted_sources_path,
        }



"""
A Memoized metaclass.
"""

import functools
from contextlib import contextmanager

from .error import TeapotError
from .log import Highlight as hl


class Memoized(type):

    """
    A Memoized type.
    """

    _ALL_INSTANCES = {}

    @classmethod
    def clear_all_instances(cls):
        for mcls in cls._ALL_INSTANCES.values():
            mcls.clear_instances()
            del mcls

        cls._ALL_INSTANCES = {}

    def __new__(cls, name, bases, attrs):

        attrs.setdefault('public_name', name)
        attrs['_INSTANCES'] = {}
        attrs['_INSTANCES_PARAMS'] = {}

        cls._ALL_INSTANCES[name] = super(Memoized, cls).__new__(cls, name, bases, attrs)
        return cls._ALL_INSTANCES[name]

    def __call__(cls, *args, **kwargs):
        keys, args, kwargs = cls.extract_keys_from_args(args, kwargs, remove_from_args=not cls.propagate_memoization_keys)
        keys = cls.transform_memoization_keys(*keys)

        if keys not in cls._INSTANCES:
            instance = super(Memoized, cls).__call__(*args, **kwargs)

            for memoization_key, key in zip(cls.memoization_keys, keys):
                setattr(instance, memoization_key, key)

            setattr(instance, 'keys', keys)

            cls._INSTANCES[keys] = instance
            cls._INSTANCES_PARAMS[keys] = (args, kwargs)

        elif cls.raise_on_duplicate_enabled:
            raise cls.DuplicateInstance(cls, keys)
        else:
            if cls.propagate_memoization_keys:
                keys, args, kwargs = cls.extract_keys_from_args(args, kwargs, remove_from_args=True)

            if (args or kwargs) and cls._INSTANCES_PARAMS[keys] != (args, kwargs):
                print (args, kwargs)
                print cls._INSTANCES_PARAMS[keys]
                raise cls.DuplicateInstance(cls, keys)

        return cls._INSTANCES[keys]


class MemoizedObject(object):
    """
    A memoized base class.
    """

    __metaclass__ = Memoized

    memoization_keys = ('name',)
    propagate_memoization_keys = False
    raise_on_duplicate_enabled = False
    duplicate_instance_message = "An instance of %s already exists with the keys %s."
    duplicate_instance_args = ('classname', 'keys')
    no_such_instance_message = "No %s could be found that matches the keys %s."
    no_such_instance_args = ('classname', 'keys')

    @classmethod
    def clear_instances(cls):
        cls._INSTANCES = {}
        cls._INSTANCES_PARAMS = {}

    @classmethod
    def transform_memoization_keys(cls, *args):
        return args

    class DuplicateInstance(TeapotError):

        """
        Another instance exist with those parameters.
        """

        def __init__(self, cls, keys):
            super(cls.DuplicateInstance, self).__init__(
                cls.get_duplicate_instance_message(keys),
                *cls.get_duplicate_instance_args(keys)
            )
            self.cls = cls
            self.keys = keys

    @classmethod
    def get_duplicate_instance_message(cls, keys):
        return cls.duplicate_instance_message

    @classmethod
    def get_duplicate_instance_args(cls, keys):
        values = {
            'classname': hl(cls.public_name),
            'keys': keys,
        }

        for memoization_key, key in zip(cls.memoization_keys, keys):
            values['key_%s' % memoization_key] = key

        return map(values.get, cls.duplicate_instance_args)

    class NoSuchInstance(TeapotError):

        """
        A non-existing instance was requested.
        """

        def __init__(self, cls, keys):
            super(cls.NoSuchInstance, self).__init__(
                cls.get_no_such_instance_message(keys),
                *cls.get_no_such_instance_args(keys)
            )
            self.cls = cls
            self.keys = keys

    @classmethod
    def get_no_such_instance_message(cls, keys):
        return cls.no_such_instance_message

    @classmethod
    def get_no_such_instance_args(cls, keys):
        values = {
            'classname': hl(cls.public_name),
            'keys': keys,
        }

        for memoization_key, key in zip(cls.memoization_keys, keys):
            values['key_%s' % memoization_key] = key

        return map(values.get, cls.no_such_instance_args)

    @classmethod
    def extract_keys_from_args(cls, args, kwargs, remove_from_args=True):
        """
        Get the keys from the specified arguments.

        If `remove_from_args` is true, then the extracted keys are
        removed from the returned `args` and `kwargs`.

        Return the keys, the `args` and the `kwargs`.
        """

        if (len(args) + len(kwargs)) < len(cls.memoization_keys):
            raise TypeError('Incorrect arguments. Got (%r, %r), expected %r' % (args, kwargs, cls.memoization_keys))

        args = list(args)
        kwargs = kwargs.copy()
        keys = []

        for index, arg in enumerate(args):
            memoization_key = cls.memoization_keys[index]

            if memoization_key in kwargs:
                raise TypeError('Got multiple values for keyword argument %r' % memoization_key)

            keys.append(arg)

        for memoization_key in cls.memoization_keys[len(args):]:
            if remove_from_args:
                keys.append(kwargs.pop(memoization_key))
            else:
                keys.append(kwargs[memoization_key])

        if remove_from_args:
            args = args[len(cls.memoization_keys):]

        return tuple(keys), tuple(args), kwargs

    @classmethod
    def get_instance(cls, *args, **kwargs):
        default = kwargs.pop('default', None)
        keys = cls.extract_keys_from_args(args, kwargs)[0]
        return cls._INSTANCES.get(keys, default)

    @classmethod
    def get_instance_or_fail(cls, *args, **kwargs):
        instance = cls.get_instance(*args, **kwargs)

        if instance is None:
            raise cls.NoSuchInstance(cls, keys=cls.extract_keys_from_args(args, kwargs)[0])

        return instance

    @classmethod
    def get_instances(cls, keys_list=None):
        if keys_list is not None:
            def to_keys(value):
                if isinstance(value, cls):
                    return value.keys

                if not isinstance(value, tuple):
                    return (value,)

                return value

            keys_list = map(to_keys, keys_list)

            result = []

            for keys in keys_list:
                if keys not in cls._INSTANCES:
                    raise cls.NoSuchInstance(cls, keys)
                result.append(cls._INSTANCES[keys])

            return result

        return cls._INSTANCES.values()

    @classmethod
    @contextmanager
    def raise_on_duplicate(cls):
        (
            sentinel,
            cls.raise_on_duplicate_enabled,
        ) = (
            cls.raise_on_duplicate_enabled,
            True,
        )

        try:
            yield
        finally:
            cls.raise_on_duplicate_enabled = sentinel

    @property
    def memoization_str(self):
        return '-'.join(map(str, (getattr(self, memoization_key) for memoization_key in self.memoization_keys)))

    def __str__(self):
        return self.memoization_str

    def __hash__(self):
        return hash(self.memoization_str)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and (other.memoization_str == self.memoization_str)
